import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class CollectionEqualsHashcodeContainsTest {

	// getReference 
	public static String getReference(Object o){
		return o.getClass().getName() + "@" + Integer.toHexString(System.identityHashCode(o));
	}

	public static void main(String[] args)  {

		// primitive type (== ڴ   true)
		int i = 1000;
		int j = 1000;
		int k = i;
		System.out.println("i == j " + (i == j));
		System.out.println("i == k " + (i == k));
		System.out.println();

		// reference type (reference  true) 
		// WARNING: JVM tries to save memory, when the Integer falls in a range (from -128 to 127). Integer v1 = 100, v2 = 100 v1 == v2 
		Object o1 = i; // boxing
		Object o2 = j; // boxing
		Object o3 = o1;
		System.out.println(getReference(o1) + " o1=" + o1); 
		System.out.println(getReference(o2) + " o2=" + o2);
		System.out.println(getReference(o3) + " o3=" + o3);
		System.out.println("o1 == o2 " + (o1 == o2)); // o1 != i2
		System.out.println("o1 == o3 " + (o1 == o3));
		System.out.println("o1 equals o2 " + o1.equals(o2));
		System.out.println("o1 equals o3 " + o1.equals(o3));
		System.out.println("o1.hashCode() == o2.hashCode() " + (o1.hashCode() == o2.hashCode()) + " hashCode=" + o1.hashCode());
		System.out.println("o1.hashCode() == o3.hashCode() " + (o1.hashCode() == o3.hashCode()));
		System.out.println();

		// String type (== ڴ reference  true)
		String s1 = "PP"; // String literal  , pool 
		String s2 = "PP"; // String literal  , pool 
		String s3 = s1;
		String s4 = "P" + "P"; // String literal  , pool 
		String s5 = new String("PP");
		String s6 = "PP" + ""; // String literal  , pool 
		String s7 = s1 + ""; // String + literal ̶  
		System.out.println(getReference(s1) + " s1=" + s1  + " hashCode=" + s1.hashCode());
		System.out.println(getReference(s2) + " s2=" + s2  + " hashCode=" + s2.hashCode());
		System.out.println(getReference(s3) + " s3=" + s3  + " hashCode=" + s3.hashCode());
		System.out.println(getReference(s4) + " s4=" + s4  + " hashCode=" + s4.hashCode());
		System.out.println(getReference(s5) + " s5=" + s5  + " hashCode=" + s5.hashCode());
		System.out.println(getReference(s6) + " s6=" + s6  + " hashCode=" + s6.hashCode());
		System.out.println(getReference(s7) + " s7=" + s7  + " hashCode=" + s7.hashCode());

		System.out.println("s1 == s2 " + (s1 == s2));
		System.out.println("s1 == s3 " + (s1 == s3));
		System.out.println("s1 == s4 " + (s1 == s4));
		System.out.println("s1 == s5 " + (s1 == s5));
		System.out.println("s1 == s6 " + (s1 == s6));
		System.out.println("s1 == s7 " + (s1 == s7));
		System.out.println("s1 equals s2 " + s1.equals(s2));
		System.out.println("s1 equals s3 " + s1.equals(s3));
		System.out.println("s1 equals s4 " + s1.equals(s4));
		System.out.println("s1 equals s5 " + s1.equals(s5));
		System.out.println("s1 equals s6 " + s1.equals(s6));
		System.out.println("s1 equals s7 " + s1.equals(s7));
		System.out.println("s1.hashCode() == s2.hashCode() " + (s1.hashCode() == s2.hashCode()));
		System.out.println("s1.hashCode() == s3.hashCode() " + (s1.hashCode() == s3.hashCode()));
		System.out.println("s1.hashCode() == s4.hashCode() " + (s1.hashCode() == s4.hashCode()));
		System.out.println("s1.hashCode() == s5.hashCode() " + (s1.hashCode() == s5.hashCode()));
		System.out.println("s1.hashCode() == s6.hashCode() " + (s1.hashCode() == s6.hashCode()));
		System.out.println("s1.hashCode() == s7.hashCode() " + (s1.hashCode() == s7.hashCode()));
		System.out.println();

		// String immutable vs. StringBuilder mutable
		// + operator (new StringBuilder(String.valueOf(str1)).append(str2).toString();
		String a = "Hello";
		String b = "Hello";
		String c = a;
		String d = "Hel" + "lo";
		String e = "Hello" + "";
		String f = a + ""; // String + StringLiteral => new String
		String g = "";
		String h = a + g; // String + String -> new String 
		String s = new String("Hello"); // new String
		String t = new String("Hello"); // new String
		String u = s + new String(""); // String + String -> new String
		String v = s + new String(""); // String + String -> new String
		String x = s.concat("~"); // concat creates new String
		String y = s.concat("~"); // concat creates new String
		System.out.println(getReference(a) + " a=" + a  + " hashCode=" + a.hashCode());
		System.out.println(getReference(b) + " b=" + b  + " hashCode=" + b.hashCode());
		System.out.println(getReference(c) + " c=" + c  + " hashCode=" + c.hashCode());
		System.out.println(getReference(d) + " d=" + d  + " hashCode=" + d.hashCode());
		System.out.println(getReference(e) + " e=" + e  + " hashCode=" + e.hashCode());
		System.out.println("a==e " + (a == e));
		System.out.println(getReference(f) + " f=" + f  + " hashCode=" + f.hashCode());
		System.out.println("a==f " + (a == f));
		System.out.println(getReference(g) + " g=" + g  + " hashCode=" + g.hashCode());
		System.out.println(getReference(h) + " h=" + h  + " hashCode=" + h.hashCode());
		System.out.println("a==h " + (a == h));
		System.out.println(getReference(s) + " s=" + s  + " hashCode=" + s.hashCode());
		System.out.println("a==s " + (a == s));
		System.out.println(getReference(t) + " t=" + t  + " hashCode=" + t.hashCode());
		System.out.println("s==t " + (s == t));
		System.out.println(getReference(u) + " u=" + u  + " hashCode=" + u.hashCode());
		System.out.println(getReference(v) + " v=" + v  + " hashCode=" + v.hashCode());
		System.out.println("u==v " + (u == v));
		System.out.println(getReference(x) + " x=" + x  + " hashCode=" + x.hashCode());
		System.out.println(getReference(y) + " y=" + y  + " hashCode=" + y.hashCode());
		System.out.println("x==y " + (x == y));

		String str1 = "P";
		String str2 = "P";
		String str3 = str1 + str2;
		String str4 = str1.concat(str2);
		System.out.println(getReference(str1) + " str1=" + str1  + " hashCode=" + str1.hashCode());
		System.out.println(getReference(str2) + " str2=" + str2  + " hashCode=" + str2.hashCode());
		System.out.println(getReference(str3) + " str3=" + str3  + " hashCode=" + str3.hashCode());
		System.out.println(getReference(str4) + " str4=" + str4  + " hashCode=" + str4.hashCode());
		System.out.println("str3==str4 " + (str3 == str4));
	
		// reference type (== ڴ reference  true)
		Person p1 = new Person("P",10, 20.5, 30.5, Gender.FEMALE);
		Person p2 = new Person("P",10, 20.5, 30.5, Gender.FEMALE);
		Person p3 = p1;
		System.out.println(getReference(p1) + " p1=" + p1);
		System.out.println(getReference(p2) + " p2=" + p2);
		System.out.println(getReference(p3) + " p3=" + p3);
		System.out.println("p1 == p2 " + (p1 == p2));
		System.out.println("p1 == p3 " + (p1 == p3));
		System.out.println("p1 equals p2 " + p1.equals(p2));
		System.out.println("p1 equals p3 " + p1.equals(p3));
		System.out.println("p1.hashCode() == p2.hashCode() " + (p1.hashCode() == p2.hashCode()) + " hashCode=" + p1.hashCode());
		System.out.println("p1.hashCode() == p3.hashCode() " + (p1.hashCode() == p3.hashCode()));
		System.out.println();

		// ArrayList
		System.out.println("pList");
		List<Person> pList = new ArrayList<Person>();
		pList.add(p1);
		pList.add(p2);
		pList.add(p3);
		pList.forEach((p) -> System.out.println(p));
		System.out.println("pList contains p1: " + pList.contains(p1));
		System.out.println("pList contains p2: " + pList.contains(p2));
		System.out.println("pList contains p3: " + pList.contains(p3));
		System.out.println("pList contains new Person: " + pList.contains(new Person("P",10, 20.5, 30.5, Gender.FEMALE)));
		System.out.println();

		// HashSet  hashCode ġϸ   Ͽ replace
		System.out.println("pSet");
		Set<Person> pSet = new HashSet<Person>();
		pSet.add(p1);
		pSet.add(p2);
		pSet.add(p3);
		pSet.forEach((p) -> System.out.println(p));
		System.out.println("pSet contains p1: " + pSet.contains(p1));
		System.out.println("pSet contains p2: " + pSet.contains(p2));
		System.out.println("pSet contains p3: " + pSet.contains(p3));
		System.out.println("pSet contains new Person: " + pSet.contains(new Person("P",10, 20.5, 30.5, Gender.FEMALE)));
		System.out.println();

		// HashMap  key  hashCode ġϸ   Ͽ replace
		System.out.println("pMap");
		Map<Person, Integer> pMap = new HashMap<Person, Integer>();
		pMap.put(p1, 1);
		pMap.put(p2, 2);
		pMap.put(p3, 3);
		pMap.forEach((p, value) -> System.out.println(p + " " + value));
		System.out.println("pMap contains p1: " + pMap.containsKey(p1));
		System.out.println("pMap contains p2: " + pMap.containsKey(p2));
		System.out.println("pMap contains p3: " + pMap.containsKey(p3));
		System.out.println("pMap contains new Person: " + pMap.containsKey(new Person("P",10, 20.5, 30.5, Gender.FEMALE)));
		System.out.println();

		// Array
		System.out.println("pArray");
		Person[] pArray = new Person[3];
		pArray[0] = p1;
		pArray[1] = p2;
		pArray[2] = p3;
		Arrays.asList(pArray).forEach(p -> System.out.println(p));
		System.out.println("pArray contains p1: " + Arrays.asList(pArray).contains(p1));
		System.out.println("pArray contains p2: " + Arrays.asList(pArray).contains(p2));
		System.out.println("pArray contains p3: " + Arrays.asList(pArray).contains(p3));
		System.out.println("pArray contains new Person: " + Arrays.asList(pArray).contains(new Person("P",10, 20.5, 30.5, Gender.FEMALE)));
		System.out.println();
    }
}
