import java.util.Comparator;
import java.util.Objects;

public class Person implements Comparable<Person> {
	protected String name; 
	protected int age;
	protected double weight;
	protected double height;
	protected Gender gender;		

	public Person() {
		this("", 0, 0.0, 0.0, Gender.FEMALE);
	}

	public Person(String name, int age, double weight, double height, Gender gender) {
		this.name = name;
		this.age = age;
		this.weight = weight;
		this.height = height;
		this.gender = gender;
	}
    
	public Person(Person other) {
		this(other.name, other.age, other.weight, other.height, other.gender);
	}

	public String getName()	{
		return name;
	}
    
	public void setName(String name) {
		this.name = name;
	}

	public int getAge() {
		return age;
	}

	public void setAge(int age) {
		this.age = age;
	}

	public double getWeight() {
		return weight;
	}

	public void setWeight(double weight) {
		this.weight = weight;
	}

	public double getHeight() {
		return height;
	}

	public void setHeight(double height) {
		this.height = height;
	}

	public Gender getGender() {
		return gender;
	}

	public void setGender(Gender gender) {
		this.gender = gender;
	}

	@Override
	public String toString() {                                        // Object.toString() overriding
		return String.format("%s,%d,%.2f,%.2f,%s", name, age, weight, height, gender.toString());
	}
   
    @Override
    public boolean equals(Object other) {                            // Object.equals overriding
		if (this == other) return true;
        if (other instanceof Person) {
          Person that = (Person) other;
          return this.getName().equals(that.getName()) && this.getAge() == that.getAge() && this.getWeight() == that.getWeight() && this.getHeight() == that.getHeight() && this.getGender().equals(that.getGender());
        }
        return false;
    }
    
    @Override 
    public int hashCode() {
        return  Objects.hash(name, age, weight, height, gender);
    }

	@Override
	public int compareTo(Person other) {
		return this.getName().compareTo(other.getName()); //ascending order
	}

	public static Comparator<Person> AgeComparator
							= new Comparator<Person>() {
		public int compare(Person p, Person q) {	      
			return p.getAge() - q.getAge(); //ascending order
		}
	};

	public static Comparator<Person> WeightComparator
							= new Comparator<Person>() {
		public int compare(Person p, Person q) {     
			return Double.compare(p.getWeight(), q.getWeight()); //ascending order
		}
	};

	public static Comparator<Person> HeightComparator
							= new Comparator<Person>() {
		public int compare(Person p, Person q) {	      
			return Double.compare(p.getHeight(), q.getHeight()); //ascending order
		}
	};

	public static Comparator<Person> GenderComparator
							= new Comparator<Person>() {
		public int compare(Person p, Person q) {	      
			return p.getGender().compareTo(q.getGender());
		}
	};

}
